/*
  Created: 03.19.05 11:34:57 by Attila Nagyidai

  $Id: C\040Console.c,v 1.1.2.1 2003/08/13 00:38:46 neum Exp $

  This file is part of IBSH (Iron Bars Shell) , a restricted Unix shell
  Copyright (C) 2005  Attila Nagyidai

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Author: Attila Nagyidai
  Email:
  URL:

*/

/* Header files */
#include "ibsh.h"

/* Code */
void Usage(char *programName)
{
  fprintf(stderr, "%s usage:\n", programName);
  /* Modify here to add your usage message when the program is
   * called without arguments */
}

/* returns the index of the first argument that is not an option; i.e.
   does not start with a dash or a slash
*/
int HandleOptions(int argc, char *argv[])
{
  int i, firstnonoption = 0;

  for (i = 1; i < argc; i++) {
    if (argv[i][0] == '/' || argv[i][0] == '-') {
      switch (argv[i][1]) {
        /* An argument -? means help is requested */
        case '?':
          Usage(argv[0]);
          break;
        case 'h':
        case 'H':
          if (!strcmp(argv[i] + 1, "help")) {
            Usage(argv[0]);
            break;
          }
          /* If the option -h means anything else
           * in your application add code here
           * Note: this falls through to the default
           * to print an "unknow option" message
          */
        /* add your option switches here */
        default:
          fprintf(stderr, "unknown option %s\n", argv[i]);
          break;
      }
    }
    else {
      firstnonoption = i;
      break;
    }
  }
  return firstnonoption;
}

/* Reads both config files, and parses the contents */
/* into arrays. */
void LoadConfig( Strng *array1, Strng *array2 )
{
  FILE *fp;
  int i = 0, j = 0;

  /* Open the commands list file first. Read every item and */
  /* store in the array. No comments please. */
  fp = fopen(COMMANDS_LIST, "r");
  if ( fp == NULL ) {
      fprintf(stderr, "Can not open file %s: %s\n", COMMANDS_LIST, strerror(errno));
      exit(0);
  }
  while ( !feof(fp) ) {
      fscanf(fp, "%s", array1[i]);
      if ( array1[i][0] != '#' ) {
          i++;
      }
  }
  fclose(fp);
  
  /* Do the same with the extensions list. */
  fp = fopen(EXTENSIONS_LIST, "r");
  if ( fp == NULL ) {
      fprintf(stderr, "Can not open file %s: %s\n", EXTENSIONS_LIST, strerror(errno));
      exit(0);
  }
  while ( !feof(fp) ) {
      fscanf(fp, "%s", array2[i]);
      if ( array2[i][0] != '#' ) {
          i++;
      }
  }
  fclose(fp);
}

/* Remove the path of the Jail root dir from the displayed paths! */
void GetPositionInJail( const char *abspath, const char *rootdir, char *relpath )
{
  int i = 0;
  int j = 0;
  
  bzero(relpath, strlen(relpath));
  for (i = strlen(rootdir); i < strlen(abspath); i++) {
      relpath[j] = abspath[i];
      j++;
  }
  relpath[j] = '\0';
}

void LTrim3( const char *base, char *result )
{
  int i = 0;
  int j = 0;

  bzero(result, strlen(result));
  for (i = 3; i < strlen(base); i++) {
      result[j] = base[i];
      j++;
  }
  result[j] = '\0';
}

int CommandOK( const char *thecommand, const char *rootdir,
const char *jailpath, char *newcommand )
{
  char *tok;
  char temp1[STRING_SIZE], *temp2;
  int i = 0, j = 0;
  int subdirlevel = -1; /* jailpath always starts with a / */
  int dirupfound = 0;
  int listed = 0;
  
  /* First, get the fancy stuff: */
  /* ../ out of the jailroot, too many ../ out of some */
  /* subdirectory in the jail, multiple commands, pipes. */
  strcpy(newcommand, "");
  
  if ( (strstr(thecommand, ";")) != NULL ) {
      return 0;
  }
  if ( (strstr(thecommand, "|")) != NULL ) {
      return 0;
  }
  if ( (strstr(thecommand, "&&")) != NULL ) {
      return 0;
  }
  if ( (strstr(thecommand, "||")) != NULL ) {
      return 0;
  }
  /* The user is in the jailroot. */
  if ( (strcmp(jailpath, "/")) == 0 ) {
      /* Does the user wish to get out ? */
      if ( (strstr(thecommand, "../")) != NULL ) {
             return 0;
      }
  }
  /* The user is deeper, than the jailroot, and */
  /* this is a problem. How deep is he, how many */
  /* ../ do we allow ?? */
  else {
      for (i = 0; i < strlen(jailpath); i++) {
          if ( jailpath[i] == '/' ) {
              subdirlevel++;
          }
      }
  }

  /* Split the command */
  for (tok = strtok((void *) thecommand, " "); tok; tok = strtok(0, " ")) {
      /* Separate parts of the command with a space */
      if ( (strlen(newcommand)) > 0 ) {
          strcat(newcommand, " ");
      }
      
      /* He wants to get to the real root, does he ? */
      /* In that case, add the jailroot to the left. */
      if ( tok[0] == '/' ) {
          strncat(newcommand, rootdir, STRING_SIZE);
      }
      
      /* how many ../ are here */
      /* if too many, that is more, then how deep */
      /* the user in the subdirs inside the jail is, */
      /* cancel the execution of the command. */
      if ( (strstr(tok, "../")) != NULL ) {
          strncpy(temp1, tok, LINE_SIZE);
          while (1) {
              temp2 = strstr(temp1, "../");
              if ( temp2 == NULL ) {
                  break;
              }
              LTrim3(temp2, temp1);
              dirupfound++;
          }
          if ( dirupfound > subdirlevel ) {
              return 0;
          }
          /* replace dirups with real path */
          for (i = 0; i < dirupfound; i++) {
              PathMinusOne(jailpath, tok, subdirlevel);
          }
      }
          /* if command is not listed, return 0 */
      i = 0;
      while ( ((strlen(commands[i])) > 0) && ( j == 0 ) ) {
          if ( (strcmp(tok, commands[i])) == 0 ) {
                listed = 1;
                break;
          }
          i++;
      }
      j++;
      strncat(newcommand, tok, STRING_SIZE);
  }
  return listed;
}

int DelBadFiles( const char *basedir )
{
  /* Check each file in the jail */
  /* and if the extensions is in the */
  /* blacklist, then erase the file. */
  /* And give a message to the user about it. */
  DIR *dp;
  struct dirent *list;
  struct stat info, attr;
  char dirtoopen[STRING_SIZE];
  int i = 0;

  strcpy(dirtoopen, basedir);
  if ( lstat( dirtoopen, &info ) == -1 ) {
      return 0;
  }

  if ( !S_ISDIR(info.st_mode) ) {
      return 0;
  }

  if ( (dp = opendir(dirtoopen)) == NULL ) {
	    closedir( dp );
      return 0;
  }

  if ( chdir(dirtoopen) == -1 ) {
      return 0;
  }

  while ( (list = readdir(dp)) != NULL ) {
      if ( (lstat(list->d_name, &attr)) == -1 ) {
          continue;
      }
      if ( S_ISDIR(attr.st_mode) ) {
          if ( ((strcmp(list->d_name, ".")) != 0) && ((strcmp(list->d_name, "..")) != 0) )
              DelBadFiles(list->d_name);
      }
      while ( (strlen(extensions[i])) > 0 ) {
          if ( (strstr(list->d_name, extensions[i])) != NULL ) {
              unlink(list->d_name);
              printf("Illegal file %s erased. Contact the sysadmin for policy.\n", list->d_name);
              break;
          }
      }
  }
  closedir( dp );
  chdir(basedir);
}

void myscanf( char *vptr )
/* wrapper for read */
{
    char c, chars[STRING_SIZE];
    int i = 0;
    
    while ( (c = getchar()) != 10 ) {
        chars[i] = c;
        i++;
    }
    chars[i] = '\0';
    strcpy(vptr, chars);
}

int  hhsystem(const char *kdozeile)   /*--- Version ohne Signalbehandlung ---*/
{
   pid_t   pid;
   int     status;

   if (kdozeile == NULL)
      return(1);   /* In Unix ist immer Kommandoprozessor vorhanden */

   if ( (pid=fork()) < 0)
      status = -1;

   else if (pid == 0) {
      execl("/bin/sh", "sh", "-c", kdozeile, NULL);
      _exit(127);

   } else
      while (waitpid(pid, &status, 0) < 0)
         if (errno != EINTR) {
            status = -1;
            break;
         }

   return(status);
}

void PathMinusOne( const char *basepath, char *evalpath, int slashcount )
{
  char *tok;
  int j = 1;
  
  bzero(evalpath, strlen(evalpath));
  if ( slashcount == 1 ) {
      strcpy(evalpath, "/");
  }
  else {
      for (tok = strtok((void *) basepath, "/"); tok; tok = strtok(0, "/")) {
          if ( j < slashcount ) {
              strncat(evalpath, tok, LINE_SIZE);
              strncat(evalpath, "/", LINE_SIZE);
          }
          j++;
      }
  }
}



/* Main: */
/* Handle arguments, read config files, start command processing. */
int main(int argc, char *argv[])
{
  char temp[STRING_SIZE], *buf;

  if (argc > 1) {
    /* If no arguments we call the Usage routine and exit */
    Usage(argv[0]);
    return 1;
  }


  /* handle the program options */
  HandleOptions(argc, argv);
  /* To Do: The code of your application goes here */
  /* First part: */
  /* Get essential information about the user who got this shell: */
  /* first the username, then the user id. Upon this, retrieve the */
  /* user's record in the passwd file. */
  bzero(&loggedin, sizeof(loggedin));
  loggedin.uid = getuid();
  loggedin.record = getpwuid(loggedin.uid);
  if ( loggedin.record == NULL ) {
      loggedin.record = getpwnam(loggedin.uname);
      if ( loggedin.record == NULL ) {
          openlog(loggedin.uname, LOG_PID, LOG_AUTH);
          syslog(LOG_ERR, "Can not obtain user information");
          closelog();
          exit(0);
      }
  }
  strncpy(loggedin.uname, loggedin.record->pw_name, STRING_SIZE);
  strncpy(loggedin.udir, loggedin.record->pw_dir, STRING_SIZE);
  /* Second part: */
  /* Handle some signal catching. Read the configuration files. */
  signal( SIGINT, SIG_IGN );
  signal( SIGQUIT, SIG_IGN );
  signal( SIGTERM, SIG_IGN );
  LoadConfig(commands, extensions);

  /* Third part: */
  /* Start reading and processing the user issued commands. */
  /* Split the command by the spaces, filter out anything, */
  /* that would allow the user to access files outside the */
  /* jail. Filter out multiples and pipes as well. No program */
  /* will be allowed to run, unless it is mentioned in the */
  /* config files. Files that are created with an extension */
  /* that is listed in the other config file, must be deleted! */
  for ( ; ; ) {
    /* Where is he ? */
    getcwd(real_path, STRING_SIZE);
    GetPositionInJail(real_path, loggedin.udir, jail_path);
    if ( (strlen(jail_path)) == 0 ) {
        strcpy(jail_path, "/");
    }
    /* We don't want the user to know where he actually is. */
    /* This is the prompt! */
    printf("[%s]%% ", jail_path);
    /* scanf("%s", user_command); */
    myscanf(user_command);
    /* Command interpretation and execution. */
    if ( (CommandOK(user_command, loggedin.udir, jail_path, filtered_command)) == 0 ) {
        printf("Sorry, can't let you do that!\n");
        continue;
    }
    /* If the user issued command starts with a shell builtin. */
    bzero(temp, strlen(temp));
    if ( (buf = strstr(filtered_command, "cd")) != NULL ) {
      if ( (strcmp(buf, filtered_command)) == 0 ) {
          LTrim3(filtered_command, temp);
          if ( (strcmp(temp, real_path)) != 0 ) {
            if ( (strcmp(temp, "..")) == 0 ) {
                PathMinusOne(jail_path, temp, 1);
            }
            if ( (strcmp(temp, "/")) == 0 ) {
                strncpy(temp, loggedin.udir, LINE_SIZE);
            }
            exitcode = chdir(temp);
            if ( exitcode == -1 ) {
                printf("ibsh: cd: %s: No such file or directory\n", temp);
            }
          }
          continue;
      }
    }
    else if ( (buf = strstr(filtered_command, "pwd")) != NULL ) {
      if ( (strcmp(buf, filtered_command)) == 0 ) {
          printf("%s\n", jail_path);
          continue;
      }
    }
    else if ( (buf = strstr(filtered_command, "logout")) != NULL ) {
      if ( (strcmp(buf, filtered_command)) == 0 ) {
        break;
      }
    }
    else if ( (buf = strstr(filtered_command, "exit")) != NULL ) {
      if ( (strcmp(buf, filtered_command)) == 0 ) {
        break;
      }
    }
    else {
        exitcode = hhsystem(filtered_command);
        if ( exitcode < 0 ) {
            printf("%s\n", strerror(errno));
        }
    }
    getcwd(real_path, STRING_SIZE);
    DelBadFiles(loggedin.udir);
    chdir(real_path);
  }
  return 0;
}

